@@ -108,7 +108,7 @@
 
 static void lapd_t200_cb(void *data);
 static void lapd_t203_cb(void *data);
-static int lapd_send_i(struct lapd_msg_ctx *lctx, int line);
+static int lapd_send_i(struct lapd_datalink *dl, int line, bool rts);
 static int lapd_est_req(struct osmo_dlsap_prim *dp, struct lapd_msg_ctx *lctx);
 
 /* UTILITY FUNCTIONS */
@@ -169,12 +169,17 @@ static void lapd_dl_flush_hist(struct lapd_datalink *dl)
 	}
 }
 
-static void lapd_dl_flush_tx(struct lapd_datalink *dl)
+static void lapd_dl_flush_tx_queue(struct lapd_datalink *dl)
 {
 	struct msgb *msg;
 
 	while ((msg = msgb_dequeue(&dl->tx_queue)))
 		msgb_free(msg);
+}
+
+static void lapd_dl_flush_tx(struct lapd_datalink *dl)
+{
+	lapd_dl_flush_tx_queue(dl);
 	lapd_dl_flush_hist(dl);
 }
 
@@ -199,11 +204,35 @@ static inline const char *lapd_state_name(enum lapd_state state)
 
 static void lapd_start_t200(struct lapd_datalink *dl)
 {
-	if (osmo_timer_pending(&dl->t200))
-		return;
-	LOGDL(dl, LOGL_INFO, "start T200 (timeout=%d.%06ds)\n",
-	      dl->t200_sec, dl->t200_usec);
-	osmo_timer_schedule(&dl->t200, dl->t200_sec, dl->t200_usec);
+	if ((dl->lapd_flags & LAPD_F_RTS)) {
+		if (dl->t200_rts != LAPD_T200_RTS_OFF)
+			return;
+		LOGDL(dl, LOGL_INFO, "Start T200. (pending until triggered by RTS)\n");
+		dl->t200_rts = LAPD_T200_RTS_PENDING;
+	} else {
+		if (osmo_timer_pending(&dl->t200))
+			return;
+		LOGDL(dl, LOGL_INFO, "Start T200 (timeout=%d.%06ds).\n", dl->t200_sec, dl->t200_usec);
+		osmo_timer_schedule(&dl->t200, dl->t200_sec, dl->t200_usec);
+	}
+}
+
+/*! Handle timeout condition of T200 in RTS mode.
+ * The caller (LAPDm code) implements the T200 timer and must detect timeout condition.
+ * The function gets called by LAPDm code when it detects a timeout of T200.
+ *  \param[in] dl caller-allocated datalink structure */
+int lapd_t200_timeout(struct lapd_datalink *dl)
+{
+	OSMO_ASSERT((dl->lapd_flags & LAPD_F_RTS));
+
+	if (dl->t200_rts != LAPD_T200_RTS_RUNNING)
+		return -EINVAL;
+
+	dl->t200_rts = LAPD_T200_RTS_OFF;
+
+	lapd_t200_cb(dl);
+
+	return 0;
 }
 
 static void lapd_start_t203(struct lapd_datalink *dl)
@@ -216,10 +245,24 @@ static void lapd_start_t203(struct lapd_datalink *dl)
 
 static void lapd_stop_t200(struct lapd_datalink *dl)
 {
-	if (!osmo_timer_pending(&dl->t200))
-		return;
+	if ((dl->lapd_flags & LAPD_F_RTS)) {
+		if (dl->t200_rts == LAPD_T200_RTS_OFF)
+			return;
+		dl->t200_rts = LAPD_T200_RTS_OFF;
+	} else {
+		if (!osmo_timer_pending(&dl->t200))
+			return;
+		osmo_timer_del(&dl->t200);
+	}
 	LOGDL(dl, LOGL_INFO, "stop T200\n");
-	osmo_timer_del(&dl->t200);
+}
+
+static bool lapd_is_t200_started(struct lapd_datalink *dl)
+{
+	if ((dl->lapd_flags & LAPD_F_RTS))
+		return (dl->t200_rts != LAPD_T200_RTS_OFF);
+	else
+		return osmo_timer_pending(&dl->t200);
 }
 
 static void lapd_stop_t203(struct lapd_datalink *dl)
@@ -354,6 +397,21 @@ void lapd_dl_reset(struct lapd_datalink *dl)
 	lapd_dl_newstate(dl, LAPD_STATE_IDLE);
 }
 
+/*! Set lapd_flags to change behaviour
+ *  \param[in] dl \ref lapd_datalink instance
+ *  \param[in] flags \ref lapd_flags */
+int lapd_dl_set_flags(struct lapd_datalink *dl, unsigned int flags)
+{
+	if (lapd_is_t200_started(dl) && (flags & LAPD_F_RTS) != (dl->lapd_flags & LAPD_F_RTS)) {
+		LOGDL(dl, LOGL_ERROR, "Changing RTS flag not allowed while T200 is running.\n");
+		return -EINVAL;
+	}
+
+	dl->lapd_flags = flags;
+
+	return 0;
+}
+
 /* reset and de-allocate history buffer */
 void lapd_dl_exit(struct lapd_datalink *dl)
 {
@@ -593,6 +651,8 @@ static void lapd_t200_cb(void *data)
 	switch (dl->state) {
 	case LAPD_STATE_SABM_SENT:
 		/* 5.4.1.3 */
+		/* increment re-transmission counter */
+		dl->retrans_ctr++;
 		if (dl->retrans_ctr >= dl->n200_est_rel + 1) {
 			/* flush tx and send buffers */
 			lapd_dl_flush_tx(dl);
@@ -611,13 +671,13 @@ static void lapd_t200_cb(void *data)
 		}
 		/* retransmit SABM command */
 		lapd_send_resend(dl);
-		/* increment re-transmission counter */
-		dl->retrans_ctr++;
 		/* restart T200 (PH-READY-TO-SEND) */
 		lapd_start_t200(dl);
 		break;
 	case LAPD_STATE_DISC_SENT:
 		/* 5.4.4.3 */
+		/* increment re-transmission counter */
+		dl->retrans_ctr++;
 		if (dl->retrans_ctr >= dl->n200_est_rel + 1) {
 			/* send MDL ERROR INIDCATION to L3 */
 			mdl_error(MDL_CAUSE_T200_EXPIRED, &dl->lctx);
@@ -635,8 +695,6 @@ static void lapd_t200_cb(void *data)
 		}
 		/* retransmit DISC command */
 		lapd_send_resend(dl);
-		/* increment re-transmission counter */
-		dl->retrans_ctr++;
 		/* restart T200 (PH-READY-TO-SEND) */
 		lapd_start_t200(dl);
 		break;
@@ -738,8 +796,9 @@ static void lapd_t203_cb(void *data)
 	lapd_start_t200(dl);
 }
 
-/* 5.5.3.1: Common function to acknowlege frames up to the given N(R) value */
-static void lapd_acknowledge(struct lapd_msg_ctx *lctx)
+/* 5.5.3.1: Common function to acknowlege frames up to the given N(R) value
+ * In case of a sequence error, the cause is returned with negative sign. */
+static int lapd_acknowledge(struct lapd_msg_ctx *lctx)
 {
 	struct lapd_datalink *dl = lctx->dl;
 	uint8_t nr = lctx->n_recv;
@@ -780,7 +839,7 @@ static void lapd_acknowledge(struct lapd_msg_ctx *lctx)
 		 */
 		if (sub_mod(nr, dl->v_ack, dl->v_range) > sub_mod(dl->v_send, dl->v_ack, dl->v_range)) {
 			LOGDL(dl, LOGL_NOTICE, "N(R) sequence error\n");
-			mdl_error(MDL_CAUSE_SEQ_ERR, lctx);
+			return -MDL_CAUSE_SEQ_ERR;
 		}
 	}
 
@@ -801,11 +860,10 @@ static void lapd_acknowledge(struct lapd_msg_ctx *lctx)
 	/* Stop T203, if running */
 	lapd_stop_t203(dl);
 	/* Start T203, if T200 is not running in MF EST state, if enabled */
-	if (!osmo_timer_pending(&dl->t200)
-	 && (dl->t203_sec || dl->t203_usec)
-	 && (dl->state == LAPD_STATE_MF_EST)) {
+	if (!lapd_is_t200_started(dl) && (dl->t203_sec || dl->t203_usec) && (dl->state == LAPD_STATE_MF_EST))
 		lapd_start_t203(dl);
-	}
+
+	return 0;
 }
 
 /* L1 -> L2 */
@@ -1242,7 +1300,7 @@ static int lapd_rx_u_ua(struct msgb *msg, struct lapd_msg_ctx *lctx)
 	/* enter multiple-frame-established state */
 	lapd_dl_newstate(dl, LAPD_STATE_MF_EST);
 	/* send outstanding frames, if any (resume / reconnect) */
-	lapd_send_i(lctx, __LINE__);
+	lapd_send_i(dl, __LINE__, false);
 	/* send notification to L3 */
 	rc = send_dl_simple(PRIM_DL_EST, PRIM_OP_CONFIRM, lctx);
 	msgb_free(msg);
@@ -1365,7 +1423,7 @@ static int lapd_rx_s(struct msgb *msg, struct lapd_msg_ctx *lctx)
 			lapd_dl_newstate(dl, LAPD_STATE_MF_EST);
 		}
 		/* Send message, if possible due to acknowledged data */
-		lapd_send_i(lctx, __LINE__);
+		lapd_send_i(dl, __LINE__, false);
 
 		break;
 	case LAPD_S_RNR:
@@ -1376,6 +1434,10 @@ static int lapd_rx_s(struct msgb *msg, struct lapd_msg_ctx *lctx)
 		/* 5.5.5 */
 		/* Set peer receiver busy condition */
 		dl->peer_busy = 1;
+		/* Flush pending messages in TX queue. */
+		lapd_dl_flush_tx_queue(dl);
+		/* stop Timer T200 */
+		lapd_stop_t200(dl);
 
 		if (lctx->p_f) {
 			if (lctx->cr == dl->cr.rem2loc.cmd) {
@@ -1402,7 +1464,7 @@ static int lapd_rx_s(struct msgb *msg, struct lapd_msg_ctx *lctx)
 			LOGDL(dl, LOGL_INFO, "RNR not polling/final state received\n");
 
 		/* Send message, if possible due to acknowledged data */
-		lapd_send_i(lctx, __LINE__);
+		lapd_send_i(dl, __LINE__, false);
 
 		break;
 	case LAPD_S_REJ:
@@ -1416,6 +1478,8 @@ static int lapd_rx_s(struct msgb *msg, struct lapd_msg_ctx *lctx)
 			dl->peer_busy = 0;
 			/* V(S) and V(A) to the N(R) in the REJ frame */
 			dl->v_send = dl->v_ack = lctx->n_recv;
+			/* Flush pending messages in TX queue. */
+			lapd_dl_flush_tx_queue(dl);
 			/* stop Timer T200 */
 			lapd_stop_t200(dl);
 			/* 5.5.3.2 */
@@ -1452,6 +1516,8 @@ static int lapd_rx_s(struct msgb *msg, struct lapd_msg_ctx *lctx)
 			dl->peer_busy = 0;
 			/* V(S) and V(A) to the N(R) in the REJ frame */
 			dl->v_send = dl->v_ack = lctx->n_recv;
+			/* Flush pending messages in TX queue. */
+			lapd_dl_flush_tx_queue(dl);
 			/* stop Timer T200 */
 			lapd_stop_t200(dl);
 			/* 5.5.7 Clear timer recovery condition */
@@ -1461,6 +1527,8 @@ static int lapd_rx_s(struct msgb *msg, struct lapd_msg_ctx *lctx)
 			dl->peer_busy = 0;
 			/* V(S) and V(A) to the N(R) in the REJ frame */
 			dl->v_send = dl->v_ack = lctx->n_recv;
+			/* Flush pending messages in TX queue. */
+			lapd_dl_flush_tx_queue(dl);
 			/* 5.5.3.2 */
 			if (lctx->cr == dl->cr.rem2loc.cmd && lctx->p_f) {
 				if (!dl->own_busy && !dl->seq_err_cond) {
@@ -1487,8 +1555,8 @@ static int lapd_rx_s(struct msgb *msg, struct lapd_msg_ctx *lctx)
 
 		/* FIXME: 5.5.4.2 2) */
 
-		/* Send message, if possible due to acknowledged data */
-		lapd_send_i(lctx, __LINE__);
+		/* Send message, if possible due to acknowledged data and new V(S) and V(A). */
+		lapd_send_i(dl, __LINE__, false);
 
 		break;
 	default:
@@ -1510,6 +1578,8 @@ static int lapd_rx_i(struct msgb *msg, struct lapd_msg_ctx *lctx)
 	uint8_t ns = lctx->n_send;
 	int length = lctx->length;
 	int rc;
+	bool i_frame_in_queue = false;
+	int mdl_cause = 0;
 
 	LOGDL(dl, LOGL_INFO, "I received in state %s on SAPI(%u)\n",
 	      lapd_state_name(dl->state), lctx->sapi);
@@ -1567,9 +1637,8 @@ static int lapd_rx_i(struct msgb *msg, struct lapd_msg_ctx *lctx)
 		      ns, dl->v_recv, lapd_state_name(dl->state));
 		/* discard data */
 		msgb_free(msg);
-		if (dl->seq_err_cond != 1) {
-			/* FIXME: help me understand what exactly todo here
-			*/
+		/* Send reject, but suppress second reject if LAPD_F_DROP_2ND_REJ flag is set. */
+		if (dl->seq_err_cond != 1 || !(dl->lapd_flags & LAPD_F_DROP_2ND_REJ)) {
 			dl->seq_err_cond = 1;
 			lapd_send_rej(lctx, lctx->p_f);
 		} else {
@@ -1587,10 +1656,12 @@ static int lapd_rx_i(struct msgb *msg, struct lapd_msg_ctx *lctx)
 		}
 		/* Even if N(s) sequence error, acknowledge to N(R)-1 */
 		/* 5.5.3.1: Acknowlege all transmitted frames up the N(R)-1 */
-		lapd_acknowledge(lctx); /* V(A) is also set here */
+		mdl_cause = lapd_acknowledge(lctx); /* V(A) is also set here */
+		if (mdl_cause < 0)
+			mdl_error(-mdl_cause, lctx);
 
 		/* Send message, if possible due to acknowledged data */
-		lapd_send_i(lctx, __LINE__);
+		lapd_send_i(dl, __LINE__, false);
 
 		return 0;
 	}
@@ -1600,8 +1671,15 @@ static int lapd_rx_i(struct msgb *msg, struct lapd_msg_ctx *lctx)
 	dl->v_recv = inc_mod(dl->v_recv, dl->v_range);
 	LOGDL(dl, LOGL_INFO, "incrementing V(R) to %u\n", dl->v_recv);
 
+	/* Update all pending frames in the queue to the new V(R) state. */
+	if (dl->update_pending_frames) {
+		rc = dl->update_pending_frames(lctx);
+		if (!rc)
+			i_frame_in_queue = true;
+	}
+
 	/* 5.5.3.1: Acknowlege all transmitted frames up the the N(R)-1 */
-	lapd_acknowledge(lctx); /* V(A) is also set here */
+	mdl_cause = lapd_acknowledge(lctx); /* V(A) is also set here */
 
 	/* Only if we are not in own receiver busy condition */
 	if (!dl->own_busy) {
@@ -1646,6 +1724,10 @@ static int lapd_rx_i(struct msgb *msg, struct lapd_msg_ctx *lctx)
 	} else
 		LOGDL(dl, LOGL_INFO, "I frame ignored during own receiver busy condition\n");
 
+	/* Indicate sequence error, if exists. */
+	if (mdl_cause < 0)
+		mdl_error(-mdl_cause, lctx);
+
 	/* Check for P bit */
 	if (lctx->p_f) {
 		/* 5.5.2.1 */
@@ -1664,14 +1746,8 @@ static int lapd_rx_i(struct msgb *msg, struct lapd_msg_ctx *lctx)
 		/* check if we are not in own receiver busy */
 		if (!dl->own_busy) {
 			/* NOTE: V(R) is already set above */
-			rc = lapd_send_i(lctx, __LINE__);
-
-			/* if update_pending_iframe returns 0 it updated
-			 * the lapd header of an iframe in the tx queue */
-			if (rc && dl->update_pending_frames)
-				rc = dl->update_pending_frames(lctx);
-
-			if (rc) {
+			rc = lapd_send_i(dl, __LINE__, false);
+			if (rc && !i_frame_in_queue) {
 				LOGDL(dl, LOGL_INFO, "we are not busy and have no pending data, "
 				      "send RR\n");
 				/* Send RR with F=0 */
@@ -1687,7 +1763,7 @@ static int lapd_rx_i(struct msgb *msg, struct lapd_msg_ctx *lctx)
 	}
 
 	/* Send message, if possible due to acknowledged data */
-	lapd_send_i(lctx, __LINE__);
+	lapd_send_i(dl, __LINE__, false);
 
 	return rc;
 }
@@ -1715,6 +1791,31 @@ int lapd_ph_data_ind(struct msgb *msg, struct lapd_msg_ctx *lctx)
 	return rc;
 }
 
+/*! Enqueue next LAPD frame and run pending T200. (Must be called when frame is ready to send.)
+ * The caller (LAPDm code) calls this function before it sends the next frame.
+ * If there is no frame in the TX queue, LAPD will enqueue next I-frame, if possible.
+ * If the T200 is pending, it is changed to running state.
+ *  \param[in] lctx LAPD context
+ *  \param[out] rc set to 1, if timer T200 state changed to running, set to 0, if not. */
+int lapd_ph_rts_ind(struct lapd_msg_ctx *lctx)
+{
+	struct lapd_datalink *dl = lctx->dl;
+
+	/* If there is no pending frame, try to enqueue next I frame. */
+	if (llist_empty(&dl->tx_queue) && (dl->state == LAPD_STATE_MF_EST || dl->state == LAPD_STATE_TIMER_RECOV)) {
+		/* Send an I frame, if there are pending outgoing messages. */
+		lapd_send_i(dl, __LINE__, true);
+	}
+
+	/* Run T200 at RTS, if pending. Tell caller that is has been started. (rc = 1) */
+	if (dl->t200_rts == LAPD_T200_RTS_PENDING) {
+		dl->t200_rts = LAPD_T200_RTS_RUNNING;
+		return 1;
+	}
+
+	return 0;
+}
+
 /* L3 -> L2 */
 
 /* send unit data */
@@ -1821,19 +1922,21 @@ static int lapd_data_req(struct osmo_dlsap_prim *dp, struct lapd_msg_ctx *lctx)
 
 	LOGDL(dl, LOGL_INFO, "writing message to send-queue: l3len: %d\n", msgb_l3len(msg));
 
+	/* Take ownership of msg, since we are keeping it around in this layer: */
+	talloc_steal(tall_lapd_ctx, msg);
 	/* Write data into the send queue */
 	msgb_enqueue(&dl->send_queue, msg);
 
 	/* Send message, if possible */
-	lapd_send_i(&dl->lctx, __LINE__);
+	lapd_send_i(dl, __LINE__, false);
 
 	return 0;
 }
 
 /* Send next I frame from queued/buffered data */
-static int lapd_send_i(struct lapd_msg_ctx *lctx, int line)
+static int lapd_send_i(struct lapd_datalink *dl, int line, bool rts)
 {
-	struct lapd_datalink *dl = lctx->dl;
+	struct lapd_msg_ctx *lctx = &dl->lctx;
 	uint8_t k = dl->k;
 	uint8_t h;
 	struct msgb *msg;
@@ -1841,18 +1944,26 @@ static int lapd_send_i(struct lapd_msg_ctx *lctx, int line)
 	int rc = - 1; /* we sent nothing */
 	struct lapd_msg_ctx nctx;
 
+	if (!rts)
+		LOGDL(dl, LOGL_INFO, "%s() called from line %d\n", __func__, line);
 
-	LOGDL(dl, LOGL_INFO, "%s() called from line %d\n", __func__, line);
+	if ((dl->lapd_flags & LAPD_F_RTS) && !llist_empty(&dl->tx_queue)) {
+		if (!rts)
+			LOGDL(dl, LOGL_INFO, "There is a frame in the TX queue, not checking for sending I frame.\n");
+		return rc;
+	}
 
 	next_frame:
 
 	if (dl->peer_busy) {
-		LOGDL(dl, LOGL_INFO, "peer busy, not sending\n");
+		if (!rts)
+			LOGDL(dl, LOGL_INFO, "Peer busy, not sending.\n");
 		return rc;
 	}
 
 	if (dl->state == LAPD_STATE_TIMER_RECOV) {
-		LOGDL(dl, LOGL_INFO, "timer recovery, not sending\n");
+		if (!rts)
+			LOGDL(dl, LOGL_INFO, "Timer recovery, not sending.\n");
 		return rc;
 	}
 
@@ -1863,8 +1974,9 @@ static int lapd_send_i(struct lapd_msg_ctx *lctx, int line)
 	 * of the error recovery procedures as described in subclauses 5.5.4 and
 	 * 5.5.7. */
 	if (dl->v_send == add_mod(dl->v_ack, k, dl->v_range)) {
-		LOGDL(dl, LOGL_INFO, "k frames outstanding, not sending more "
-		      "(k=%u V(S)=%u V(A)=%u)\n", k, dl->v_send, dl->v_ack);
+		if (!rts)
+			LOGDL(dl, LOGL_INFO, "k frames outstanding, not sending more. (k=%u V(S)=%u V(A)=%u)\n",
+			      k, dl->v_send, dl->v_ack);
 		return rc;
 	}
 
@@ -1906,7 +2018,7 @@ static int lapd_send_i(struct lapd_msg_ctx *lctx, int line)
 		msg = lapd_msgb_alloc(length, "LAPD I");
 		msg->l3h = msgb_put(msg, length);
 		/* assemble message */
-		memcpy(&nctx, &dl->lctx, sizeof(nctx));
+		memcpy(&nctx, lctx, sizeof(nctx));
 		/* keep nctx.ldp */
 		/* keep nctx.sapi */
 		/* keep nctx.tei */
@@ -1936,7 +2048,7 @@ static int lapd_send_i(struct lapd_msg_ctx *lctx, int line)
 		msg = lapd_msgb_alloc(length, "LAPD I resend");
 		msg->l3h = msgb_put(msg, length);
 		/* assemble message */
-		memcpy(&nctx, &dl->lctx, sizeof(nctx));
+		memcpy(&nctx, lctx, sizeof(nctx));
 		/* keep nctx.ldp */
 		/* keep nctx.sapi */
 		/* keep nctx.tei */
@@ -1958,7 +2070,7 @@ static int lapd_send_i(struct lapd_msg_ctx *lctx, int line)
 	/* If timer T200 is not running at the time right before transmitting a
 	 * frame, when the PH-READY-TO-SEND primitive is received from the
 	 * physical layer., it shall be set. */
-	if (!osmo_timer_pending(&dl->t200)) {
+	if (!lapd_is_t200_started(dl)) {
 		/* stop Timer T203, if running */
 		lapd_stop_t203(dl);
 		/* start Timer T200 */
@@ -1967,7 +2079,11 @@ static int lapd_send_i(struct lapd_msg_ctx *lctx, int line)
 
 	dl->send_ph_data_req(&nctx, msg);
 
-	rc = 0; /* we sent something */
+	/* When using RTS, we send only one frame. */
+	if ((dl->lapd_flags & LAPD_F_RTS))
+		return 0;
+
+	rc = 0; /* We sent an I frame, so sending RR frame is not required. */
 	goto next_frame;
 }
 
